#!/usr/bin/env perl
#
#	@(#) security.pl v3.1.2 System Security script
#

use Getopt::Std;
use File::Basename;
use Sys::Hostname;
use POSIX;
use Digest::SHA1;
use Net::SMTP;
#use strict;
use warnings;
no warnings 'once';

our $CONFDIR = "/.security";
our $OWNER = 0;
our $GROUP = 0;
our $file_open = 0;
our $host = hostname();
our ($smtp, $title, @modlist, @mailhost, %chksum, $format, @load, $opts, %opts);

sub usage {
	my $PROGNAME = File::Basename::basename($0);

	print "Usage: $PROGNAME -[$opts]\n";
	print "\t-C <conf file>\tUse <conf file> instead of default config file.\n";
	print "\t-d\t\tDebug mode\n";
	print "\t-h\t\tThis help screen\n";
	print "\t-i\t\tIgnore missing config files\n";
	print "\t-M <addr>\tMail report to <addr>, defaults to 'root\@$host'\n";
	print "\t\t\tif <addr> starts with a /, write to a file.\n";
	exit 0;
}

sub SHA1 {
	my ($file) = @_;
	my ($sha) = new Digest::SHA1;
	local *sha_fd;

	open (sha_fd, $file) || return "SHA1($file) = -1\n";
	$sha->reset();
	$sha->addfile(*sha_fd{IO});
	my ($hex) = $sha->hexdigest();
	close sha_fd;
	return "SHA1($file) = $hex\n";
}

sub init {
	print ( STDERR "In init\n") if ( $opts{'d'} );

	&check_conf_files();
	&read_config('load');
	@load = ( $^O, @load );		# Get the OS specific module loaded too
	&load_modules();
	&report_header();
}

sub check_conf_files {
	$format = "SHA1\\((.*)\\) = (.*)";
	print ( STDERR "In check_conf_files\n") if ( $opts{'d'} );

	$opts{'C'} = "$CONFDIR/security.cf" if ( ! defined $opts{'C'} );

	# Check the integerty of the config directory
	&check_conf_file ($CONFDIR, 00700, $OWNER, $GROUP, 'config dir');

	# Check the integerty of the modules directory
	&check_conf_file ("$CONFDIR/Security", 00700, $OWNER, $GROUP, 'modules dir');

	# Now check the digest file to make sure all the other files are ok.
	# This idea came from the FreeBSD ports collection
	&check_conf_file ("$CONFDIR/.digest", 00600, $OWNER, $GROUP, 'digest file');

	# Now read in the digest data
	open (FP, "$CONFDIR/.digest") || die "Could not open digest file\n";
	while (<FP>) {
		grep (/^$format$/ && ($chksum{$1}=$2), $_);
	}
	close FP;
	unlink ( "$CONFDIR/.digest" );

	# Now check if the configfile is ok.
	&check_conf_file ("$opts{'C'}", 00400, $OWNER, $GROUP, 'config file');

	# Check if this script itself is ok
	&check_conf_file ($0, 00700, $OWNER, $GROUP, 'security script');

	# Check if the checksum db is ok
	&check_conf_file ("$CONFDIR/.checksum.db", 00600, $OWNER, $GROUP, 'checksum db');

	# Now check all the installed modules
	@modlist = &module_list();
	for (my ($i) = 0; $i<=$#modlist; $i++) {
		&check_conf_file($modlist[$i], 00600, $OWNER, $GROUP, "module file $modlist[$i]");
	}
}

sub read_config {
	my ($field) = @_;
	local *FP;

	print ( STDERR "In read_config\n") if ( $opts{'d'} );
	open (FP, "$opts{'C'}") || die "Could not read config file ($opts{'C'})\n";
	while (<FP>) {
		chop;
		@{$field} = split (/[ \t]+/o, substr ($_, index ($_, ' ')+1))
			if (/^$field /);
	}

	for (my ($i) = 0; $i <= $#{$field}; $i++) {
		printf STDERR "\t$field = ${$field}[$i]\n" if ( $opts{'d'} );
	}
	close FP;
}

sub load_modules {
	print ( STDERR "In load_modules\n") if ( $opts{'d'} );

	# Set the new directory for loading the specific modules
	@INC = ($CONFDIR, @INC);

	# Now load all the requested modules from the config file.
	for (my ($i) = 0; $i <= $#load; $i++) {
		printf STDERR "\tLoading Security::$load[$i]\n" if ($opts{'d'});
		my ($pkg) = "use Security::$load[$i]";
		eval $pkg;
	}
}

sub report_header {
	my ($date, $version);
	local *FP;

	# Mail report information
	$date = strftime ( "%m/%d/%Y", localtime(time()));
	open (FP, $0) || die "Could not open myself $0\n";
	grep (/@\(#\) .[^ ]* ([^ ]*)/ && ($version=$1), <FP>);
	close FP;
	$title = "$host security ($version) report for $date";
}

sub parse_options {
	$opts = 'C:dhiM:';
	&getopts($opts, \%opts);

	&usage() if ( $opts{'h'} );
}

sub call {
	my ($func) = @_;
	my ($cmd);
	printf STDERR "In call($func)\n" if ( $opts{'d'} );

	for (my ($i) = 0; $i <= $#load; $i++) {
		$cmd = "Security::$load[$i]::$func";
		&{$cmd} if ( defined &{$cmd} );
	}
}

sub create_secure_file {
	my ($file, $uid, $gid, $mode) = @_;

	umask 077;
	$( = $uid; $) = $gid;
	open (FP, "> $file") || die "Could not open secure file\n";
	close FP;
	chown ( $uid, $gid, $file);
	chmod ( $mode, $file);
}

sub check_conf_file {
	print ( STDERR "In check_conf_file\n") if ( $opts{'d'} );
	my ($file, $wanted_mode, $wanted_uid, $wanted_gid, $msg) = @_;
	my ($cursum);

	if (! -e $file ) {
		die "$msg does not exist\n" if ( ! $opts{'i'} );
		return;
	}

    	my($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($file);

	printf STDERR "\t%s mode = %o, uid = %d, gid = %d\n", 
		$file, ($mode & 00777), $uid, $gid if ( $opts{'d'} );

	die "Permissions, owner or group not correct on $msg\n" 
		if (($mode & 00777) != $wanted_mode || 
			$uid != $wanted_uid || 
			$gid != $GROUP );

	if ( exists $chksum{$file} && length ($chksum{$file}) > 1 ) {

		grep (/^$format$/ && ($cursum=$2), &SHA1($file));

		printf STDERR "\t$file checksum <%s> actual <%s> expected\n", 
			$cursum, $chksum{$file} if ( $opts{'d'} );

		die "$msg has bad checksum\n" if ( $chksum{$file} ne $cursum );
	}
}

sub module_list {
	local *FP;

	# Get the list of modules that are installed
	opendir (FP, "$CONFDIR/Security") || die "Could not read modules dir\n";
	my (@a) = readdir FP;
	for (my ($i, $j)=(0,0); $i <= $#a; $i++) {
		$modlist[$j++] = "$CONFDIR/Security/$a[$i]" 
			if ( -f "$CONFDIR/Security/$a[$i]" );
	}
	closedir FP;

	return @modlist;
}

sub update_files {
	local *FP;
	my ($file, $sum, $i, $j);
	my (@files) = ( $0, "$CONFDIR/security.cf", "$CONFDIR/.checksum.db" );

	print ( STDERR "In update_files\n") if ( $opts{'d'} );

	@files = ( @files, &module_list());

	&create_secure_file ("$CONFDIR/.digest", $OWNER, $GROUP, 00600);
	open (FP, "> $CONFDIR/.digest") || die "Could not open digest file";

	for $file (@files) {
		$sum = &SHA1($file);
		print ( STDERR "\tSum = $sum" ) if ( $opts{'d'} );
		print FP $sum || die "Error saving digest digest";
	}
	close FP;

}

sub check_fp {
	$opts{'M'} = "root@$host" if ( ! defined $opts{'M'} );

	if ( !$file_open ) {
		if ( $opts{'M'} =~ /^\//o ) {
			&create_secure_file ($opts{'M'}, $OWNER, $GROUP, 00600);
			open (OUTPUT, ">$opts{'M'}" ) 
				|| die "Could not open $opts{'M'}";
			print (OUTPUT "$title\n");
			$file_open = 1;
		} else { 
			&read_config('mailhost');
			$smtp = Net::SMTP->new($mailhost[0]);
			$smtp->mail("root\@$host");
			$smtp->to($opts{'M'});
			$smtp->data();
			$smtp->datasend("To: $opts{'M'}\n");
			$smtp->datasend("From: root\@$host\n");
			$smtp->datasend("Subject: $title\n");
			$smtp->datasend("\n");

			$file_open = 2;
		}
	}
}

sub print_data {
	print ( STDERR "In print_data\n") if ( $opts{'d'} );
	my ($header,%ary) = @_;
	my ($cnt) = 0;
	foreach my $i (sort keys %ary) {
		&check_fp();
		if (! $cnt ) {
			if ($file_open == 1) {
				print OUTPUT $header;
			} else {
				$smtp->datasend($header);
			}
			$cnt=1;
		}
		if ( $file_open == 1) {
			print (OUTPUT $i);
			print (OUTPUT " $ary{$i}") if $ary{$i} ne '1';
			print (OUTPUT "\n");
		} else {
			$smtp->datasend($i);
			$smtp->datasend(" $ary{$i}") if $ary{$i} ne '1';
			$smtp->datasend("\n");
		}
	}
}

# Main.
&parse_options();
&init();
&call('init');
&call('check');
&call('print');
&call('destroy');
if ($file_open == 2) {
	$smtp->dataend();
	$smtp->quit;
}
&update_files();

exit(0);
